@@ -1,17 +1,18 @@
 // Implementation of the function builtin.
 #include "config.h"  // IWYU pragma: keep
 
-#include <errno.h>
-#include <stddef.h>
-#include <stdlib.h>
+#include "builtin_function.h"
+
 #include <unistd.h>
 
+#include <cerrno>
+#include <cstddef>
+#include <cstdlib>
 #include <memory>
 #include <string>
 #include <vector>
 
 #include "builtin.h"
-#include "builtin_function.h"
 #include "common.h"
 #include "complete.h"
 #include "event.h"
@@ -29,35 +30,48 @@ struct function_cmd_opts_t {
     bool print_help = false;
     bool shadow_scope = true;
     wcstring description = L"";
-    std::vector<event_t> events;
+    std::vector<event_description_t> events;
     wcstring_list_t named_arguments;
     wcstring_list_t inherit_vars;
     wcstring_list_t wrap_targets;
 };
 
-// This command is atypical in using the "+" (REQUIRE_ORDER) option for flag parsing.
+// This command is atypical in using the "-" (RETURN_IN_ORDER) option for flag parsing.
 // This is needed due to the semantics of the -a/--argument-names flag.
-static const wchar_t *const short_options = L"+:a:d:e:hj:p:s:v:w:SV:";
-static const struct woption long_options[] = {{L"description", required_argument, NULL, 'd'},
-                                              {L"on-signal", required_argument, NULL, 's'},
-                                              {L"on-job-exit", required_argument, NULL, 'j'},
-                                              {L"on-process-exit", required_argument, NULL, 'p'},
-                                              {L"on-variable", required_argument, NULL, 'v'},
-                                              {L"on-event", required_argument, NULL, 'e'},
-                                              {L"wraps", required_argument, NULL, 'w'},
-                                              {L"help", no_argument, NULL, 'h'},
-                                              {L"argument-names", required_argument, NULL, 'a'},
-                                              {L"no-scope-shadowing", no_argument, NULL, 'S'},
-                                              {L"inherit-variable", required_argument, NULL, 'V'},
-                                              {NULL, 0, NULL, 0}};
+static const wchar_t *const short_options = L"-:a:d:e:hj:p:s:v:w:SV:";
+static const struct woption long_options[] = {
+    {L"description", required_argument, nullptr, 'd'},
+    {L"on-signal", required_argument, nullptr, 's'},
+    {L"on-job-exit", required_argument, nullptr, 'j'},
+    {L"on-process-exit", required_argument, nullptr, 'p'},
+    {L"on-variable", required_argument, nullptr, 'v'},
+    {L"on-event", required_argument, nullptr, 'e'},
+    {L"wraps", required_argument, nullptr, 'w'},
+    {L"help", no_argument, nullptr, 'h'},
+    {L"argument-names", required_argument, nullptr, 'a'},
+    {L"no-scope-shadowing", no_argument, nullptr, 'S'},
+    {L"inherit-variable", required_argument, nullptr, 'V'},
+    {nullptr, 0, nullptr, 0}};
 
 static int parse_cmd_opts(function_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncss method)
                           int argc, wchar_t **argv, parser_t &parser, io_streams_t &streams) {
     const wchar_t *cmd = L"function";
     int opt;
     wgetopter_t w;
-    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
+    bool handling_named_arguments = false;
+    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
+        if (opt != 'a' && opt != 1) handling_named_arguments = false;
         switch (opt) {
+            case 1: {
+                if (handling_named_arguments) {
+                    opts.named_arguments.push_back(w.woptarg);
+                    break;
+                } else {
+                    streams.err.append_format(_(L"%ls: Unexpected positional argument '%ls'"), cmd,
+                                              w.woptarg);
+                    return STATUS_INVALID_ARGS;
+                }
+            }
             case 'd': {
                 opts.description = w.woptarg;
                 break;
@@ -68,7 +82,7 @@ static int parse_cmd_opts(function_cmd_opts_t &opts, int *optind,  //!OCLINT(hig
                     streams.err.append_format(_(L"%ls: Unknown signal '%ls'"), cmd, w.woptarg);
                     return STATUS_INVALID_ARGS;
                 }
-                opts.events.push_back(event_t::signal_event(sig));
+                opts.events.push_back(event_description_t::signal(sig));
                 break;
             }
             case 'v': {
@@ -77,35 +91,23 @@ static int parse_cmd_opts(function_cmd_opts_t &opts, int *optind,  //!OCLINT(hig
                     return STATUS_INVALID_ARGS;
                 }
 
-                opts.events.push_back(event_t::variable_event(w.woptarg));
+                opts.events.push_back(event_description_t::variable(w.woptarg));
                 break;
             }
             case 'e': {
-                opts.events.push_back(event_t::generic_event(w.woptarg));
+                opts.events.push_back(event_description_t::generic(w.woptarg));
                 break;
             }
             case 'j':
             case 'p': {
                 pid_t pid;
-                event_t e(EVENT_ANY);
+                event_description_t e(event_type_t::any);
 
                 if ((opt == 'j') && (wcscasecmp(w.woptarg, L"caller") == 0)) {
                     job_id_t job_id = -1;
 
-                    if (is_subshell) {
-                        size_t block_idx = 0;
-
-                        // Find the outermost substitution block.
-                        for (block_idx = 0;; block_idx++) {
-                            const block_t *b = parser.block_at_index(block_idx);
-                            if (b == NULL || b->type() == SUBST) break;
-                        }
-
-                        // Go one step beyond that, to get to the caller.
-                        const block_t *caller_block = parser.block_at_index(block_idx + 1);
-                        if (caller_block != NULL && caller_block->job != NULL) {
-                            job_id = caller_block->job->job_id;
-                        }
+                    if (parser.libdata().is_subshell) {
+                        job_id = parser.libdata().caller_job_id;
                     }
 
                     if (job_id == -1) {
@@ -113,11 +115,11 @@ static int parse_cmd_opts(function_cmd_opts_t &opts, int *optind,  //!OCLINT(hig
                             _(L"%ls: Cannot find calling job for event handler"), cmd);
                         return STATUS_INVALID_ARGS;
                     }
-                    e.type = EVENT_JOB_ID;
+                    e.type = event_type_t::job_exit;
                     e.param1.job_id = job_id;
                 } else if ((opt == 'p') && (wcscasecmp(w.woptarg, L"%self") == 0)) {
                     pid = getpid();
-                    e.type = EVENT_EXIT;
+                    e.type = event_type_t::exit;
                     e.param1.pid = pid;
                 } else {
                     pid = fish_wcstoi(w.woptarg);
@@ -127,13 +129,14 @@ static int parse_cmd_opts(function_cmd_opts_t &opts, int *optind,  //!OCLINT(hig
                         return STATUS_INVALID_ARGS;
                     }
 
-                    e.type = EVENT_EXIT;
+                    e.type = event_type_t::exit;
                     e.param1.pid = (opt == 'j' ? -1 : 1) * abs(pid);
                 }
                 opts.events.push_back(e);
                 break;
             }
             case 'a': {
+                handling_named_arguments = true;
                 opts.named_arguments.push_back(w.woptarg);
                 break;
             }
@@ -204,7 +207,8 @@ static int validate_function_name(int argc, const wchar_t *const *argv, wcstring
 /// Define a function. Calls into `function.cpp` to perform the heavy lifting of defining a
 /// function.
 int builtin_function(parser_t &parser, io_streams_t &streams, const wcstring_list_t &c_args,
-                     const parsed_source_ref_t &source, tnode_t<grammar::job_list> body) {
+                     const parsed_source_ref_t &source,
+                     tnode_t<grammar::block_statement> func_node) {
     assert(source && "Missing source in builtin_function");
     // The wgetopt function expects 'function' as the first argument. Make a new wcstring_list with
     // that property. This is needed because this builtin has a different signature than the other
@@ -230,13 +234,17 @@ int builtin_function(parser_t &parser, io_streams_t &streams, const wcstring_lis
     if (retval != STATUS_CMD_OK) return retval;
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_CMD_OK;
     }
 
     if (argc != optind) {
-        if (opts.named_arguments.size()) {
+        if (!opts.named_arguments.empty()) {
             for (int i = optind; i < argc; i++) {
+                if (!valid_var_name(argv[i])) {
+                    streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, argv[i]);
+                    return STATUS_INVALID_ARGS;
+                }
                 opts.named_arguments.push_back(argv[i]);
             }
         } else {
@@ -247,23 +255,26 @@ int builtin_function(parser_t &parser, io_streams_t &streams, const wcstring_lis
     }
 
     // We have what we need to actually define the function.
-    function_data_t d;
-    d.name = function_name;
-    if (!opts.description.empty()) d.description = opts.description;
-    // d.description = opts.description;
-    d.events.swap(opts.events);
-    d.props.shadow_scope = opts.shadow_scope;
-    d.props.named_arguments = std::move(opts.named_arguments);
-    d.inherit_vars = std::move(opts.inherit_vars);
+    auto props = std::make_shared<function_properties_t>();
+    props->shadow_scope = opts.shadow_scope;
+    props->named_arguments = std::move(opts.named_arguments);
+    props->parsed_source = source;
+    props->func_node = func_node;
 
-    for (size_t i = 0; i < d.events.size(); i++) {
-        event_t &e = d.events.at(i);
-        e.function_name = d.name;
+    // Populate inherit_vars.
+    for (const wcstring &name : opts.inherit_vars) {
+        if (auto var = parser.vars().get(name)) {
+            props->inherit_vars[name] = var->as_list();
+        }
     }
 
-    d.props.parsed_source = source;
-    d.props.body_node = body;
-    function_add(std::move(d), parser);
+    // Add the function itself.
+    function_add(function_name, opts.description, props, parser.libdata().current_filename);
+
+    // Add any event handlers.
+    for (const event_description_t &ed : opts.events) {
+        event_add_handler(std::make_shared<event_handler_t>(ed, function_name));
+    }
 
     // Handle wrap targets by creating the appropriate completions.
     for (const wcstring &wt : opts.wrap_targets) complete_add_wrapper(function_name, wt);
