@@ -2,9 +2,8 @@
 #ifndef FISH_BUILTIN_JOBS_H
 #define FISH_BUILTIN_JOBS_H
 
-#include <wchar.h>
-
 #include <cstring>
+#include <cwchar>
 
 class parser_t;
 
