@@ -1,23 +1,24 @@
 // Implementation of the read builtin.
 #include "config.h"  // IWYU pragma: keep
 
-#include <errno.h>
-#include <limits.h>
-#include <stddef.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
+#include "builtin_read.h"
+
 #include <unistd.h>
-#include <wchar.h>
 
 #include <algorithm>
+#include <cerrno>
+#include <climits>
+#include <cstddef>
+#include <cstdio>
+#include <cstdlib>
+#include <cstring>
+#include <cwchar>
 #include <memory>
 #include <numeric>
 #include <string>
 #include <vector>
 
 #include "builtin.h"
-#include "builtin_read.h"
 #include "common.h"
 #include "complete.h"
 #include "env.h"
@@ -26,6 +27,7 @@
 #include "highlight.h"
 #include "history.h"
 #include "io.h"
+#include "parser.h"
 #include "proc.h"
 #include "reader.h"
 #include "wcstringutil.h"
@@ -36,14 +38,15 @@ struct read_cmd_opts_t {
     bool print_help = false;
     int place = ENV_USER;
     wcstring prompt_cmd;
-    const wchar_t *prompt = NULL;
-    const wchar_t *prompt_str = NULL;
+    const wchar_t *prompt = nullptr;
+    const wchar_t *prompt_str = nullptr;
     const wchar_t *right_prompt = L"";
     const wchar_t *commandline = L"";
     // If a delimiter was given. Used to distinguish between the default
     // empty string and a given empty delimiter.
     bool have_delimiter = false;
     wcstring delimiter;
+    bool tokenize = false;
     bool shell = false;
     bool array = false;
     bool silent = false;
@@ -53,35 +56,34 @@ struct read_cmd_opts_t {
     bool one_line = false;
 };
 
-static const wchar_t *const short_options = L":ac:d:ghiLlm:n:p:suxzP:UR:LB";
-static const struct woption long_options[] = {
-    {L"array", no_argument, NULL, 'a'},
-    {L"command", required_argument, NULL, 'c'},
-    {L"delimiter", required_argument, NULL, 'd'},
-    {L"export", no_argument, NULL, 'x'},
-    {L"global", no_argument, NULL, 'g'},
-    {L"help", no_argument, NULL, 'h'},
-    {L"line", no_argument, NULL, 'L'},
-    {L"local", no_argument, NULL, 'l'},
-    {L"mode-name", required_argument, NULL, 'm'},
-    {L"nchars", required_argument, NULL, 'n'},
-    {L"null", no_argument, NULL, 'z'},
-    {L"prompt", required_argument, NULL, 'p'},
-    {L"prompt-str", required_argument, NULL, 'P'},
-    {L"right-prompt", required_argument, NULL, 'R'},
-    {L"shell", no_argument, NULL, 'S'},
-    {L"silent", no_argument, NULL, 's'},
-    {L"unexport", no_argument, NULL, 'u'},
-    {L"universal", no_argument, NULL, 'U'},
-    {NULL, 0, NULL, 0}
-};
+static const wchar_t *const short_options = L":ac:d:ghiLlm:n:p:sStuxzP:UR:LB";
+static const struct woption long_options[] = {{L"array", no_argument, nullptr, 'a'},
+                                              {L"command", required_argument, nullptr, 'c'},
+                                              {L"delimiter", required_argument, nullptr, 'd'},
+                                              {L"export", no_argument, nullptr, 'x'},
+                                              {L"global", no_argument, nullptr, 'g'},
+                                              {L"help", no_argument, nullptr, 'h'},
+                                              {L"line", no_argument, nullptr, 'L'},
+                                              {L"list", no_argument, nullptr, 'a'},
+                                              {L"local", no_argument, nullptr, 'l'},
+                                              {L"nchars", required_argument, nullptr, 'n'},
+                                              {L"null", no_argument, nullptr, 'z'},
+                                              {L"prompt", required_argument, nullptr, 'p'},
+                                              {L"prompt-str", required_argument, nullptr, 'P'},
+                                              {L"right-prompt", required_argument, nullptr, 'R'},
+                                              {L"shell", no_argument, nullptr, 'S'},
+                                              {L"silent", no_argument, nullptr, 's'},
+                                              {L"tokenize", no_argument, nullptr, 't'},
+                                              {L"unexport", no_argument, nullptr, 'u'},
+                                              {L"universal", no_argument, nullptr, 'U'},
+                                              {nullptr, 0, nullptr, 0}};
 
 static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncss method)
                           int argc, wchar_t **argv, parser_t &parser, io_streams_t &streams) {
     wchar_t *cmd = argv[0];
     int opt;
     wgetopter_t w;
-    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
+    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
         switch (opt) {
             case 'a': {
                 opts.array = true;
@@ -97,8 +99,9 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
                 break;
             }
             case 'i': {
-                streams.err.append_format(_(L"%ls: usage of -i for --silent is deprecated. Please use -s or --silent instead.\n"),
-                        cmd);
+                streams.err.append_format(_(L"%ls: usage of -i for --silent is deprecated. Please "
+                                            L"use -s or --silent instead.\n"),
+                                          cmd);
                 return STATUS_INVALID_ARGS;
             }
             case L'g': {
@@ -117,25 +120,18 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
                 opts.place |= ENV_LOCAL;
                 break;
             }
-            case L'm': {
-                streams.err.append_format(_(L"%ls: flags '--mode-name' / '-m' are now ignored. "
-                                            L"Set fish_history instead.\n"),
-                                          cmd);
-                break;
-            }
             case L'n': {
                 opts.nchars = fish_wcstoi(w.woptarg);
                 if (errno) {
                     if (errno == ERANGE) {
                         streams.err.append_format(_(L"%ls: Argument '%ls' is out of range\n"), cmd,
                                                   w.woptarg);
-                        builtin_print_help(parser, streams, cmd, streams.err);
+                        builtin_print_error_trailer(parser, streams.err, cmd);
                         return STATUS_INVALID_ARGS;
                     }
 
-                    streams.err.append_format(_(L"%ls: Argument '%ls' must be an integer\n"), cmd,
-                                              w.woptarg);
-                    builtin_print_help(parser, streams, cmd, streams.err);
+                    streams.err.append_format(BUILTIN_ERR_NOT_NUMBER, cmd, w.woptarg);
+                    builtin_print_error_trailer(parser, streams.err, cmd);
                     return STATUS_INVALID_ARGS;
                 }
                 break;
@@ -160,6 +156,10 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
                 opts.shell = true;
                 break;
             }
+            case L't': {
+                opts.tokenize = true;
+                break;
+            }
             case L'U': {
                 opts.place |= ENV_UNIVERSAL;
                 break;
@@ -197,20 +197,18 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
 
 /// Read from the tty. This is only valid when the stream is stdin and it is attached to a tty and
 /// we weren't asked to split on null characters.
-static int read_interactive(wcstring &buff, int nchars, bool shell, bool silent,
+static int read_interactive(parser_t &parser, wcstring &buff, int nchars, bool shell, bool silent,
                             const wchar_t *prompt, const wchar_t *right_prompt,
                             const wchar_t *commandline) {
     int exit_res = STATUS_CMD_OK;
-    const wchar_t *line;
 
-    wcstring read_history_ID = history_session_id();
-    if (!read_history_ID.empty()) read_history_ID += L"_read";
-    reader_push(read_history_ID);
+    // Don't keep history
+    reader_push(parser, L"");
 
     reader_set_left_prompt(prompt);
     reader_set_right_prompt(right_prompt);
     if (shell) {
-        reader_set_complete_function(&complete);
+        reader_set_complete_ok(true);
         reader_set_highlight_function(&highlight_shell);
         reader_set_test_function(&reader_shell_test);
     }
@@ -220,21 +218,20 @@ static int read_interactive(wcstring &buff, int nchars, bool shell, bool silent,
     reader_set_exit_on_interrupt(true);
     reader_set_silent_status(silent);
 
-    reader_set_buffer(commandline, wcslen(commandline));
-    proc_push_interactive(1);
+    reader_set_buffer(commandline, std::wcslen(commandline));
+    scoped_push<bool> interactive{&parser.libdata().is_interactive, true};
 
-    event_fire_generic(L"fish_prompt");
-    line = reader_readline(nchars);
-    proc_pop_interactive();
-    if (line) {
-        if (0 < nchars && (size_t)nchars < wcslen(line)) {
+    event_fire_generic(parser, L"fish_prompt");
+    auto mline = reader_readline(nchars);
+    interactive.restore();
+    if (mline) {
+        buff = mline.acquire();
+        if (nchars > 0 && static_cast<size_t>(nchars) < buff.size()) {
             // Line may be longer than nchars if a keybinding used `commandline -i`
             // note: we're deliberately throwing away the tail of the commandline.
             // It shouldn't be unread because it was produced with `commandline -i`,
             // not typed.
-            buff = wcstring(line, nchars);
-        } else {
-            buff = wcstring(line);
+            buff.resize(nchars);
         }
     } else {
         exit_res = STATUS_CMD_ERROR;
@@ -275,7 +272,10 @@ static int read_in_chunks(int fd, wcstring &buff, bool split_null) {
         if (bytes_consumed < bytes_read) {
             // We found a splitter. The +1 because we need to treat the splitter as consumed, but
             // not append it to the string.
-            CHECK(lseek(fd, bytes_consumed - bytes_read + 1, SEEK_CUR) != -1, STATUS_CMD_ERROR)
+            if (lseek(fd, bytes_consumed - bytes_read + 1, SEEK_CUR) == -1) {
+                wperror(L"lseek");
+                return STATUS_CMD_ERROR;
+            }
             finished = true;
         } else if (str.size() > read_byte_limit) {
             exit_res = STATUS_READ_TOO_MUCH;
@@ -313,13 +313,13 @@ static int read_one_char_at_a_time(int fd, wcstring &buff, int nchars, bool spli
 
             nbytes++;
             if (MB_CUR_MAX == 1) {
-                res = (unsigned char)b;
+                res = static_cast<unsigned char>(b);
                 finished = true;
             } else {
-                size_t sz = mbrtowc(&res, &b, 1, &state);
-                if (sz == (size_t)-1) {
-                    memset(&state, 0, sizeof(state));
-                } else if (sz != (size_t)-2) {
+                size_t sz = std::mbrtowc(&res, &b, 1, &state);
+                if (sz == static_cast<size_t>(-1)) {
+                    std::memset(&state, 0, sizeof(state));
+                } else if (sz != static_cast<size_t>(-2)) {
                     finished = true;
                 }
             }
@@ -334,7 +334,7 @@ static int read_one_char_at_a_time(int fd, wcstring &buff, int nchars, bool spli
         if (split_null && res == L'\0') break;
 
         buff.push_back(res);
-        if (nchars > 0 && (size_t)nchars <= buff.size()) {
+        if (nchars > 0 && static_cast<size_t>(nchars) <= buff.size()) {
             break;
         }
     }
@@ -350,17 +350,20 @@ static int read_one_char_at_a_time(int fd, wcstring &buff, int nchars, bool spli
 static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int argc,
                               const wchar_t *const *argv, parser_t &parser, io_streams_t &streams) {
     if (opts.prompt && opts.prompt_str) {
-        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"-p", L"-P");
-        builtin_print_help(parser, streams, cmd, streams.err);
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd,
+                                  L"-p", L"-P");
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     if (opts.have_delimiter && opts.one_line) {
-        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"--delimiter", L"--line");
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd,
+                                  L"--delimiter", L"--line");
         return STATUS_INVALID_ARGS;
     }
     if (opts.one_line && opts.split_null) {
-        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"-z", L"--line");
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd,
+                                  L"-z", L"--line");
         return STATUS_INVALID_ARGS;
     }
 
@@ -373,7 +376,7 @@ static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int arg
 
     if ((opts.place & ENV_UNEXPORT) && (opts.place & ENV_EXPORT)) {
         streams.err.append_format(BUILTIN_ERR_EXPUNEXP, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -381,7 +384,7 @@ static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int arg
             (opts.place & ENV_UNIVERSAL ? 1 : 0) >
         1) {
         streams.err.append_format(BUILTIN_ERR_GLOCAL, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -400,11 +403,23 @@ static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int arg
         return STATUS_INVALID_ARGS;
     }
 
+    if (opts.tokenize && opts.have_delimiter) {
+        streams.err.append_format(BUILTIN_ERR_COMBO2, cmd,
+                                  L"--delimiter and --tokenize can not be used together");
+        return STATUS_INVALID_ARGS;
+    }
+
+    if (opts.tokenize && opts.one_line) {
+        streams.err.append_format(BUILTIN_ERR_COMBO2, cmd,
+                                  L"--line and --tokenize can not be used together");
+        return STATUS_INVALID_ARGS;
+    }
+
     // Verify all variable names.
     for (int i = 0; i < argc; i++) {
         if (!valid_var_name(argv[i])) {
             streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, argv[i]);
-            builtin_print_help(parser, streams, cmd, streams.err);
+            builtin_print_error_trailer(parser, streams.err, cmd);
             return STATUS_INVALID_ARGS;
         }
     }
@@ -414,6 +429,7 @@ static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int arg
 
 /// The read builtin. Reads from stdin and stores the values in environment variables.
 int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+    auto &vars = parser.vars();
     wchar_t *cmd = argv[0];
     int argc = builtin_count_args(argv);
     wcstring buff;
@@ -433,7 +449,7 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     }
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
+        builtin_print_help(parser, streams, cmd);
         return STATUS_CMD_OK;
     }
 
@@ -448,33 +464,34 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         opts.shell = false;
     }
 
-    wchar_t * const *var_ptr = argv;
-    auto vars_left = [&] () { return argv + argc - var_ptr; };
-    auto clear_remaining_vars = [&] () {
+    wchar_t *const *var_ptr = argv;
+    auto vars_left = [&]() { return argv + argc - var_ptr; };
+    auto clear_remaining_vars = [&]() {
         while (vars_left()) {
-            env_set_empty(*var_ptr, opts.place);
-            // env_set_one(*var_ptr, opts.place, L"");
+            parser.vars().set_empty(*var_ptr, opts.place);
             ++var_ptr;
         }
     };
 
-    // Normally, we either consume a line of input or all available input. But if we are reading a line at
-    // a time, we need a middle ground where we only consume as many lines as we need to fill the given vars.
+    // Normally, we either consume a line of input or all available input. But if we are reading a
+    // line at a time, we need a middle ground where we only consume as many lines as we need to
+    // fill the given vars.
     do {
         buff.clear();
 
-        // TODO: Determine if the original set of conditions for interactive reads should be reinstated:
-        // if (isatty(0) && streams.stdin_fd == STDIN_FILENO && !split_null) {
+        // TODO: Determine if the original set of conditions for interactive reads should be
+        // reinstated: if (isatty(0) && streams.stdin_fd == STDIN_FILENO && !split_null) {
         int stream_stdin_is_a_tty = isatty(streams.stdin_fd);
         if (stream_stdin_is_a_tty && !opts.split_null) {
             // Read interactively using reader_readline(). This does not support splitting on null.
-            exit_res = read_interactive(buff, opts.nchars, opts.shell, opts.silent, opts.prompt,
-                                        opts.right_prompt, opts.commandline);
+            exit_res = read_interactive(parser, buff, opts.nchars, opts.shell, opts.silent,
+                                        opts.prompt, opts.right_prompt, opts.commandline);
         } else if (!opts.nchars && !stream_stdin_is_a_tty &&
                    lseek(streams.stdin_fd, 0, SEEK_CUR) != -1) {
             exit_res = read_in_chunks(streams.stdin_fd, buff, opts.split_null);
         } else {
-            exit_res = read_one_char_at_a_time(streams.stdin_fd, buff, opts.nchars, opts.split_null);
+            exit_res =
+                read_one_char_at_a_time(streams.stdin_fd, buff, opts.nchars, opts.split_null);
         }
 
         if (exit_res != STATUS_CMD_OK) {
@@ -487,19 +504,57 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             return exit_res;
         }
 
+        if (opts.tokenize) {
+            tokenizer_t tok{buff.c_str(), TOK_ACCEPT_UNFINISHED};
+            wcstring out;
+            if (opts.array) {
+                // Array mode: assign each token as a separate element of the sole var.
+                wcstring_list_t tokens;
+                while (auto t = tok.next()) {
+                    auto text = tok.text_of(*t);
+                    if (unescape_string(text, &out, UNESCAPE_DEFAULT)) {
+                        tokens.push_back(out);
+                    } else {
+                        tokens.push_back(text);
+                    }
+                }
+
+                vars.set(*var_ptr++, opts.place, tokens);
+            } else {
+                maybe_t<tok_t> t;
+                while ((vars_left() - 1 > 0) && (t = tok.next())) {
+                    auto text = tok.text_of(*t);
+                    if (unescape_string(text, &out, UNESCAPE_DEFAULT)) {
+                        vars.set_one(*var_ptr++, opts.place, out);
+                    } else {
+                        vars.set_one(*var_ptr++, opts.place, text);
+                    }
+                }
+
+                // If we still have tokens, set the last variable to them.
+                if ((t = tok.next())) {
+                    wcstring rest = wcstring(buff, t->offset);
+                    vars.set_one(*var_ptr++, opts.place, rest);
+                }
+            }
+            // The rest of the loop is other split-modes, we don't care about those.
+            continue;
+        }
+
         if (!opts.have_delimiter) {
-            auto ifs = env_get(L"IFS");
+            auto ifs = parser.vars().get(L"IFS");
             if (!ifs.missing_or_empty()) opts.delimiter = ifs->as_string();
         }
 
         if (opts.delimiter.empty()) {
-            // Every character is a separate token with one wrinkle involving non-array mode where the
-            // final var gets the remaining characters as a single string.
+            // Every character is a separate token with one wrinkle involving non-array mode where
+            // the final var gets the remaining characters as a single string.
             size_t x = std::max(static_cast<size_t>(1), buff.size());
-            size_t n_splits = (opts.array || static_cast<size_t>(vars_left()) > x) ? x : vars_left();
+            size_t n_splits =
+                (opts.array || static_cast<size_t>(vars_left()) > x) ? x : vars_left();
             wcstring_list_t chars;
             chars.reserve(n_splits);
-            x = x - n_splits + 1;
+
             int i = 0;
             for (auto it = buff.begin(), end = buff.end(); it != end; ++i, ++it) {
                 if (opts.array || i + 1 < vars_left()) {
@@ -512,19 +567,18 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
 
             if (opts.array) {
                 // Array mode: assign each char as a separate element of the sole var.
-                env_set(*var_ptr++, opts.place, chars);
+                vars.set(*var_ptr++, opts.place, chars);
             } else {
-                // Not array mode: assign each char to a separate var with the remainder being assigned
-                // to the last var.
-                auto c = chars.begin();
-                for (; c != chars.end() && vars_left(); ++c) {
-                    env_set_one(*var_ptr++, opts.place, *c);
+                // Not array mode: assign each char to a separate var with the remainder being
+                // assigned to the last var.
+                for (const auto &c : chars) {
+                    vars.set_one(*var_ptr++, opts.place, c);
                 }
             }
         } else if (opts.array) {
-            // The user has requested the input be split into a sequence of tokens and all the tokens
-            // assigned to a single var. How we do the tokenizing depends on whether the user specified
-            // the delimiter string or we're using IFS.
+            // The user has requested the input be split into a sequence of tokens and all the
+            // tokens assigned to a single var. How we do the tokenizing depends on whether the user
+            // specified the delimiter string or we're using IFS.
             if (!opts.have_delimiter) {
                 // We're using IFS, so tokenize the buffer using each IFS char. This is for backward
                 // compatibility with old versions of fish.
@@ -534,14 +588,14 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                      loc.first != wcstring::npos; loc = wcstring_tok(buff, opts.delimiter, loc)) {
                     tokens.emplace_back(wcstring(buff, loc.first, loc.second));
                 }
-                env_set(*var_ptr++, opts.place, tokens);
+                vars.set(*var_ptr++, opts.place, tokens);
             } else {
                 // We're using a delimiter provided by the user so use the `string split` behavior.
                 wcstring_list_t splits;
                 split_about(buff.begin(), buff.end(), opts.delimiter.begin(), opts.delimiter.end(),
                             &splits);
 
-                env_set(*var_ptr++, opts.place, splits);
+                vars.set(*var_ptr++, opts.place, splits);
             }
         } else {
             // Not array mode. Split the input into tokens and assign each to the vars in sequence.
@@ -553,9 +607,17 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                     wcstring substr;
                     loc = wcstring_tok(buff, (vars_left() > 1) ? opts.delimiter : wcstring(), loc);
                     if (loc.first != wcstring::npos) {
+                        if (vars_left() == 1) {  // Discard trailing delimiters, see #6406
+                            loc.first =
+                                std::find_if(buff.begin() + loc.first, buff.end(),
+                                             [&opts](wchar_t c) {
+                                                 return opts.delimiter.find(c) == wcstring::npos;
+                                             }) -
+                                buff.begin();
+                        }
                         substr = wcstring(buff, loc.first, loc.second);
                     }
-                    env_set_one(*var_ptr++, opts.place, substr);
+                    vars.set_one(*var_ptr++, opts.place, substr);
                 }
             } else {
                 // We're using a delimiter provided by the user so use the `string split` behavior.
@@ -564,9 +626,9 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                 // is set to the remaining string.
                 split_about(buff.begin(), buff.end(), opts.delimiter.begin(), opts.delimiter.end(),
                             &splits, argc - 1);
-                assert(splits.size() <= (size_t) vars_left());
+                assert(splits.size() <= (size_t)vars_left());
                 for (const auto &split : splits) {
-                    env_set_one(*var_ptr++, opts.place, split);
+                    vars.set_one(*var_ptr++, opts.place, split);
                 }
             }
         }
