@@ -1,13 +1,15 @@
 // Implementation of the source builtin.
 #include "config.h"  // IWYU pragma: keep
 
+#include "builtin_source.h"
+
 #include <fcntl.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include <wchar.h>
+
+#include <cwchar>
 
 #include "builtin.h"
-#include "builtin_source.h"
 #include "common.h"
 #include "env.h"
 #include "fallback.h"  // IWYU pragma: keep
@@ -31,7 +33,7 @@ int builtin_source(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     if (retval != STATUS_CMD_OK) return retval;
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
+        builtin_print_help(parser, streams, cmd);
         return STATUS_CMD_OK;
     }
 
@@ -39,7 +41,7 @@ int builtin_source(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     struct stat buf;
     const wchar_t *fn, *fn_intern;
 
-    if (argc == optind || wcscmp(argv[optind], L"-") == 0) {
+    if (argc == optind || std::wcscmp(argv[optind], L"-") == 0) {
         // Either a bare `source` which means to implicitly read from stdin or an explicit `-`.
         if (argc == optind && isatty(streams.stdin_fd)) {
             // Don't implicitly read from the terminal.
@@ -73,14 +75,17 @@ int builtin_source(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         fn_intern = intern(argv[optind]);
     }
 
-    const source_block_t *sb = parser.push_block<source_block_t>(fn_intern);
-    reader_push_current_filename(fn_intern);
+    const block_t *sb = parser.push_block(block_t::source_block(fn_intern));
+    auto &ld = parser.libdata();
+    scoped_push<const wchar_t *> filename_push{&ld.current_filename, fn_intern};
 
     // This is slightly subtle. If this is a bare `source` with no args then `argv + optind` already
     // points to the end of argv. Otherwise we want to skip the file name to get to the args if any.
-    env_set_argv(argv + optind + (argc == optind ? 0 : 1));
+    wcstring_list_t argv_list =
+        null_terminated_array_t<wchar_t>::to_list(argv + optind + (argc == optind ? 0 : 1));
+    parser.vars().set_argv(std::move(argv_list));
 
-    retval = reader_read(fd, streams.io_chain ? *streams.io_chain : io_chain_t());
+    retval = reader_read(parser, fd, streams.io_chain ? *streams.io_chain : io_chain_t());
 
     parser.pop_block(sb);
 
@@ -88,11 +93,10 @@ int builtin_source(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         streams.err.append_format(_(L"%ls: Error while reading file '%ls'\n"), cmd,
                                   fn_intern == intern_static(L"-") ? L"<stdin>" : fn_intern);
     } else {
-        retval = proc_get_last_status();
+        retval = parser.get_last_status();
     }
 
     // Do not close fd after calling reader_read. reader_read automatically closes it before calling
     // eval.
-    reader_pop_current_filename();
     return retval;
 }
