@@ -1,17 +1,18 @@
 // Color class implementation.
 #include "config.h"  // IWYU pragma: keep
 
-#include <stddef.h>
-#include <stdint.h>
-#include <stdlib.h>
-#include <wchar.h>  // IWYU pragma: keep
-
 #include "color.h"
+
+#include <cstddef>
+#include <cstdint>
+#include <cstdlib>
+#include <cwchar>  // IWYU pragma: keep
+
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
 
 bool rgb_color_t::try_parse_special(const wcstring &special) {
-    memset(&data, 0, sizeof data);
+    std::memset(&data, 0, sizeof data);
     const wchar_t *name = special.c_str();
     if (!wcscasecmp(name, L"normal")) {
         this->type = type_normal;
@@ -79,20 +80,22 @@ static int parse_hex_digit(wchar_t x) {
         case L'F': {
             return 0xF;
         }
-        default: { return -1; }
+        default: {
+            return -1;
+        }
     }
 }
 
 static unsigned long squared_difference(long p1, long p2) {
-    unsigned long diff = (unsigned long)labs(p1 - p2);
+    unsigned long diff = static_cast<unsigned long>(labs(p1 - p2));
     return diff * diff;
 }
 
 static unsigned char convert_color(const unsigned char rgb[3], const uint32_t *colors,
                                    size_t color_count) {
     long r = rgb[0], g = rgb[1], b = rgb[2];
-    unsigned long best_distance = (unsigned long)-1;
-    unsigned char best_index = (unsigned char)-1;
+    unsigned long best_distance = static_cast<unsigned long>(-1);
+    unsigned char best_index = static_cast<unsigned char>(-1);
     for (unsigned char idx = 0; idx < color_count; idx++) {
         uint32_t color = colors[idx];
         long test_r = (color >> 16) & 0xFF, test_g = (color >> 8) & 0xFF,
@@ -108,7 +111,7 @@ static unsigned char convert_color(const unsigned char rgb[3], const uint32_t *c
 }
 
 bool rgb_color_t::try_parse_rgb(const wcstring &name) {
-    memset(&data, 0, sizeof data);
+    std::memset(&data, 0, sizeof data);
     // We support the following style of rgb formats (case insensitive):
     //  #FA3
     //  #F3A035
@@ -171,7 +174,7 @@ wcstring_list_t rgb_color_t::named_color_names() {
     wcstring_list_t result;
     result.reserve(1 + count);
     for (size_t i = 0; i < count; i++) {
-        if (named_colors[i].hidden == false) {
+        if (!named_colors[i].hidden) {
             result.push_back(named_colors[i].name);
         }
     }
@@ -185,7 +188,7 @@ wcstring_list_t rgb_color_t::named_color_names() {
 }
 
 bool rgb_color_t::try_parse_named(const wcstring &str) {
-    memset(&data, 0, sizeof data);
+    std::memset(&data, 0, sizeof data);
     size_t max = sizeof named_colors / sizeof *named_colors;
     for (size_t idx = 0; idx < max; idx++) {
         if (0 == wcscasecmp(str.c_str(), named_colors[idx].name)) {
@@ -290,7 +293,7 @@ unsigned char rgb_color_t::to_name_index() const {
     assert(type == type_named || type == type_rgb);
     if (type == type_named) return data.name_idx;
     if (type == type_rgb) return term16_color_for_rgb(data.color.rgb);
-    return (unsigned char)-1;  // this is an error
+    return static_cast<unsigned char>(-1);  // this is an error
 }
 
 void rgb_color_t::parse(const wcstring &str) {
@@ -299,7 +302,7 @@ void rgb_color_t::parse(const wcstring &str) {
     if (!success) success = try_parse_named(str);
     if (!success) success = try_parse_rgb(str);
     if (!success) {
-        memset(&this->data, 0, sizeof this->data);
+        std::memset(&this->data, 0, sizeof this->data);
         this->type = type_none;
     }
 }
@@ -316,7 +319,7 @@ wcstring rgb_color_t::description() const {
             return L"none";
         }
         case type_named: {
-            return format_string(L"named(%d: %ls)", (int)data.name_idx,
+            return format_string(L"named(%d: %ls)", static_cast<int>(data.name_idx),
                                  name_for_color_idx(data.name_idx));
         }
         case type_rgb: {
@@ -329,7 +332,9 @@ wcstring rgb_color_t::description() const {
         case type_normal: {
             return L"normal";
         }
-        default: { break; }
+        default: {
+            break;
+        }
     }
     DIE("unknown color type");
 }
