@@ -2,6 +2,8 @@
 #ifndef FISH_IOTHREAD_H
 #define FISH_IOTHREAD_H
 
+#include <pthread.h>
+
 #include <functional>
 #include <type_traits>
 
@@ -11,7 +13,7 @@
 /// pointer, and returns an int, which is passed to the completionCallback.
 /// \param completionCallback The function to execute on the main thread once the background thread
 /// is complete. Accepts an int (the return value of handler) and the context.
-/// \param context A arbitary context pointer to pass to the handler and completion callback.
+/// \param context An arbitrary context pointer to pass to the handler and completion callback.
 /// \return A sequence number, currently not very useful.
 int iothread_perform_base(int (*handler)(void *), void (*completionCallback)(void *, int),
                           void *context);
@@ -21,14 +23,16 @@ int iothread_perform_base(int (*handler)(void *), void (*completionCallback)(voi
 /// \return A file descriptor on which to listen for completion callbacks.
 int iothread_port(void);
 
-/// Services one iothread competion callback.
+/// Services one iothread completion callback.
 void iothread_service_completion(void);
 
 /// Waits for all iothreads to terminate.
-void iothread_drain_all(void);
+/// \return the number of threads that were running.
+int iothread_drain_all(void);
 
 // Internal implementation
-int iothread_perform_impl(std::function<void(void)> &&func, std::function<void(void)> &&completion);
+int iothread_perform_impl(std::function<void(void)> &&func, std::function<void(void)> &&completion,
+                          bool cant_wait = false);
 
 // Template helpers
 // This is the glue part of the handler-completion handoff
@@ -69,10 +73,28 @@ int iothread_perform(const HANDLER &handler, const COMPLETION &completion) {
 
 // variant of iothread_perform without a completion handler
 inline int iothread_perform(std::function<void(void)> &&func) {
-    return iothread_perform_impl(std::move(func), std::function<void(void)>());
+    return iothread_perform_impl(std::move(func), {});
+}
+
+/// Variant of iothread_perform that disrespects the thread limit.
+/// It does its best to spawn a new thread if all other threads are occupied.
+/// This is for cases where deferring a new thread might lead to deadlock.
+inline int iothread_perform_cantwait(std::function<void(void)> &&func) {
+    return iothread_perform_impl(std::move(func), {}, true);
 }
 
 /// Performs a function on the main thread, blocking until it completes.
 void iothread_perform_on_main(std::function<void(void)> &&func);
 
+/// Creates a pthread, manipulating the signal mask so that the thread receives no signals.
+/// The thread is detached.
+/// The pthread runs \p func.
+/// \returns true on success, false on failure.
+bool make_detached_pthread(void *(*func)(void *), void *param);
+bool make_detached_pthread(std::function<void(void)> &&func);
+
+/// \returns a thread ID for this thread.
+/// Thread IDs are never repeated.
+uint64_t thread_id();
+
 #endif
