@@ -7,11 +7,11 @@
 #include <fcntl.h>
 #include <libgen.h>
 #include <limits.h>
-#include <pthread.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
 #include <sys/stat.h>
+
+#include <cstring>
 #if defined(__linux__)
 #include <sys/statfs.h>
 #endif
@@ -19,19 +19,27 @@
 #include <sys/statvfs.h>
 #include <sys/types.h>
 #include <unistd.h>
-#include <wchar.h>
 #include <wctype.h>
 
+#include <atomic>
+#include <cwchar>
 #include <string>
 #include <unordered_map>
 
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
-#include "wutil.h"     // IWYU pragma: keep
+#include "flog.h"
+#include "wutil.h"  // IWYU pragma: keep
 
-typedef std::string cstring;
+using cstring = std::string;
 
-const file_id_t kInvalidFileID = {(dev_t)-1LL, (ino_t)-1LL, (uint64_t)-1LL, -1, -1, -1, -1};
+const file_id_t kInvalidFileID = {static_cast<dev_t>(-1LL),
+                                  static_cast<ino_t>(-1LL),
+                                  static_cast<uint64_t>(-1LL),
+                                  -1,
+                                  -1,
+                                  -1,
+                                  -1};
 
 /// Map used as cache by wgettext.
 static owning_lock<std::unordered_map<wcstring, wcstring>> wgettext_map;
@@ -39,7 +47,7 @@ static owning_lock<std::unordered_map<wcstring, wcstring>> wgettext_map;
 bool wreaddir_resolving(DIR *dir, const wcstring &dir_path, wcstring &out_name, bool *out_is_dir) {
     struct dirent *result = readdir(dir);
     if (!result) {
-        out_name = L"";
+        out_name.clear();
         return false;
     }
 
@@ -85,7 +93,7 @@ bool wreaddir_resolving(DIR *dir, const wcstring &dir_path, wcstring &out_name,
 bool wreaddir(DIR *dir, wcstring &out_name) {
     struct dirent *result = readdir(dir);
     if (!result) {
-        out_name = L"";
+        out_name.clear();
         return false;
     }
 
@@ -94,7 +102,7 @@ bool wreaddir(DIR *dir, wcstring &out_name) {
 }
 
 bool wreaddir_for_dirs(DIR *dir, wcstring *out_name) {
-    struct dirent *result = NULL;
+    struct dirent *result = nullptr;
     while (!result) {
         result = readdir(dir);
         if (!result) break;
@@ -119,26 +127,20 @@ bool wreaddir_for_dirs(DIR *dir, wcstring *out_name) {
     if (result && out_name) {
         *out_name = str2wcstring(result->d_name);
     }
-    if (!result) return false;
-    return true;
+    return result != nullptr;
 }
 
-const wcstring wgetcwd() {
+wcstring wgetcwd() {
     char cwd[PATH_MAX];
     char *res = getcwd(cwd, sizeof(cwd));
     if (res) {
         return str2wcstring(res);
     }
 
-    debug(0, _(L"getcwd() failed with errno %d/%s"), errno, strerror(errno));
+    FLOGF(error, _(L"getcwd() failed with errno %d/%s"), errno, std::strerror(errno));
     return wcstring();
 }
 
-int wchdir(const wcstring &dir) {
-    cstring tmp = wcs2string(dir);
-    return chdir(tmp.c_str());
-}
-
 FILE *wfopen(const wcstring &path, const char *mode) {
     int permissions = 0, options = 0;
     size_t idx = 0;
@@ -159,7 +161,7 @@ FILE *wfopen(const wcstring &path, const char *mode) {
         }
         default: {
             errno = EINVAL;
-            return NULL;
+            return nullptr;
         }
     }
     // Skip binary.
@@ -169,37 +171,45 @@ FILE *wfopen(const wcstring &path, const char *mode) {
     if (mode[idx] == '+') permissions = O_RDWR;
 
     int fd = wopen_cloexec(path, permissions | options, 0666);
-    if (fd < 0) return NULL;
+    if (fd < 0) return nullptr;
     FILE *result = fdopen(fd, mode);
-    if (result == NULL) close(fd);
+    if (result == nullptr) close(fd);
     return result;
 }
 
-bool set_cloexec(int fd) {
+int set_cloexec(int fd, bool should_set) {
+    // Note we don't want to overwrite existing flags like O_NONBLOCK which may be set. So fetch the
+    // existing flags and modify them.
     int flags = fcntl(fd, F_GETFD, 0);
     if (flags < 0) {
-        return false;
+        return -1;
     }
-    if (flags & FD_CLOEXEC) {
-        return true;
+    int new_flags = flags;
+    if (should_set) {
+        new_flags |= FD_CLOEXEC;
+    } else {
+        new_flags &= ~FD_CLOEXEC;
+    }
+    if (flags == new_flags) {
+        return 0;
+    } else {
+        return fcntl(fd, F_SETFD, new_flags);
     }
-    return fcntl(fd, F_SETFD, flags | FD_CLOEXEC) >= 0;
 }
 
-static int wopen_internal(const wcstring &pathname, int flags, mode_t mode, bool cloexec) {
+int open_cloexec(const std::string &cstring, int flags, mode_t mode, bool cloexec) {
     ASSERT_IS_NOT_FORKED_CHILD();
-    cstring tmp = wcs2string(pathname);
     int fd;
 
 #ifdef O_CLOEXEC
     // Prefer to use O_CLOEXEC. It has to both be defined and nonzero.
     if (cloexec) {
-        fd = open(tmp.c_str(), flags | O_CLOEXEC, mode);
+        fd = open(cstring.c_str(), flags | O_CLOEXEC, mode);
     } else {
-        fd = open(tmp.c_str(), flags, mode);
+        fd = open(cstring.c_str(), flags, mode);
     }
 #else
-    fd = open(tmp.c_str(), flags, mode);
+    fd = open(cstring.c_str(), flags, mode);
     if (fd >= 0 && !set_cloexec(fd)) {
         close(fd);
         fd = -1;
@@ -208,8 +218,14 @@ static int wopen_internal(const wcstring &pathname, int flags, mode_t mode, bool
     return fd;
 }
 
+int wopen(const wcstring &pathname, int flags, mode_t mode) {
+    cstring tmp = wcs2string(pathname);
+    return open(tmp.c_str(), flags, mode);
+}
+
 int wopen_cloexec(const wcstring &pathname, int flags, mode_t mode) {
-    return wopen_internal(pathname, flags, mode, true);
+    cstring tmp = wcs2string(pathname);
+    return open_cloexec(tmp, flags, mode, true);
 }
 
 DIR *wopendir(const wcstring &name) {
@@ -229,13 +245,9 @@ dir_t::~dir_t() {
     }
 }
 
-bool dir_t::valid() const {
-    return this->dir != nullptr;
-}
+bool dir_t::valid() const { return this->dir != nullptr; }
 
-bool dir_t::read(wcstring &name) {
-    return wreaddir(this->dir, name);
-}
+bool dir_t::read(wcstring &name) { return wreaddir(this->dir, name); }
 
 int wstat(const wcstring &file_name, struct stat *buf) {
     const cstring tmp = wcs2string(file_name);
@@ -260,9 +272,9 @@ int wunlink(const wcstring &file_name) {
 void wperror(const wchar_t *s) {
     int e = errno;
     if (s[0] != L'\0') {
-        fwprintf(stderr, L"%ls: ", s);
+        std::fwprintf(stderr, L"%ls: ", s);
     }
-    fwprintf(stderr, L"%s\n", strerror(e));
+    std::fwprintf(stderr, L"%s\n", std::strerror(e));
 }
 
 int make_fd_nonblocking(int fd) {
@@ -287,7 +299,7 @@ int make_fd_blocking(int fd) {
 
 int fd_check_is_remote(int fd) {
 #if defined(__linux__)
-    struct statfs buf{0};
+    struct statfs buf {};
     if (fstatfs(fd, &buf) < 0) {
         return -1;
     }
@@ -302,6 +314,12 @@ int fd_check_is_remote(int fd) {
             // Other FSes are assumed local.
             return 0;
     }
+#elif defined(ST_LOCAL)
+    // ST_LOCAL is a flag to statvfs, which is itself standardized.
+    // In practice the only system to use this path is NetBSD.
+    struct statvfs buf {};
+    if (fstatvfs(fd, &buf) < 0) return -1;
+    return (buf.f_flag & ST_LOCAL) ? 0 : 1;
 #elif defined(MNT_LOCAL)
     struct statfs buf {};
     if (fstatfs(fd, &buf) < 0) return -1;
@@ -312,7 +330,7 @@ int fd_check_is_remote(int fd) {
 }
 
 static inline void safe_append(char *buffer, const char *s, size_t buffsize) {
-    strncat(buffer, s, buffsize - strlen(buffer) - 1);
+    std::strncat(buffer, s, buffsize - std::strlen(buffer) - 1);
 }
 
 // In general, strerror is not async-safe, and therefore we cannot use it directly. So instead we
@@ -322,10 +340,10 @@ const char *safe_strerror(int err) {
 #if defined(__UCLIBC__)
     // uClibc does not have sys_errlist, however, its strerror is believed to be async-safe.
     // See issue #808.
-    return strerror(err);
+    return std::strerror(err);
 #elif defined(HAVE__SYS__ERRS) || defined(HAVE_SYS_ERRLIST)
 #ifdef HAVE_SYS_ERRLIST
-    if (err >= 0 && err < sys_nerr && sys_errlist[err] != NULL) {
+    if (err >= 0 && err < sys_nerr && sys_errlist[err] != nullptr) {
         return sys_errlist[err];
     }
 #elif defined(HAVE__SYS__ERRS)
@@ -367,7 +385,7 @@ void safe_perror(const char *message) {
     safe_append(buff, safe_strerror(err), sizeof buff);
     safe_append(buff, "\n", sizeof buff);
 
-    ignore_result(write(STDERR_FILENO, buff, strlen(buff)));
+    ignore_result(write(STDERR_FILENO, buff, std::strlen(buff)));
     errno = err;
 }
 
@@ -526,8 +544,8 @@ static void wgettext_really_init() {
 /// For wgettext: Internal init function. Automatically called when a translation is first
 /// requested.
 static void wgettext_init_if_necessary() {
-    static pthread_once_t once = PTHREAD_ONCE_INIT;
-    pthread_once(&once, wgettext_really_init);
+    static std::once_flag s_wgettext_init{};
+    std::call_once(s_wgettext_init, wgettext_really_init);
 }
 
 const wcstring &wgettext(const wchar_t *in) {
@@ -598,7 +616,7 @@ int fish_iswgraph(wint_t wc) {
 /// Convenience variants on fish_wcwswidth().
 ///
 /// See fallback.h for the normal definitions.
-int fish_wcswidth(const wchar_t *str) { return fish_wcswidth(str, wcslen(str)); }
+int fish_wcswidth(const wchar_t *str) { return fish_wcswidth(str, std::wcslen(str)); }
 
 /// Convenience variants on fish_wcwswidth().
 ///
@@ -606,7 +624,7 @@ int fish_wcswidth(const wchar_t *str) { return fish_wcswidth(str, wcslen(str));
 int fish_wcswidth(const wcstring &str) { return fish_wcswidth(str.c_str(), str.size()); }
 
 locale_t fish_c_locale() {
-    static locale_t loc = newlocale(LC_ALL_MASK, "C", NULL);
+    static const locale_t loc = newlocale(LC_ALL_MASK, "C", nullptr);
     return loc;
 }
 
@@ -629,7 +647,7 @@ int fish_wcstoi(const wchar_t *str, const wchar_t **endptr, int base) {
 
     errno = 0;
     wchar_t *_endptr;
-    long result = wcstol(str, &_endptr, base);
+    long result = std::wcstol(str, &_endptr, base);
     if (result > INT_MAX) {
         result = INT_MAX;
         errno = ERANGE;
@@ -646,7 +664,7 @@ int fish_wcstoi(const wchar_t *str, const wchar_t **endptr, int base) {
         }
     }
     if (endptr) *endptr = _endptr;
-    return (int)result;
+    return static_cast<int>(result);
 }
 
 /// An enhanced version of wcstol().
@@ -667,7 +685,7 @@ long fish_wcstol(const wchar_t *str, const wchar_t **endptr, int base) {
 
     errno = 0;
     wchar_t *_endptr;
-    long result = wcstol(str, &_endptr, base);
+    long result = std::wcstol(str, &_endptr, base);
     while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
     if (!errno && *_endptr) {
         if (_endptr == str) {
@@ -698,7 +716,7 @@ long long fish_wcstoll(const wchar_t *str, const wchar_t **endptr, int base) {
 
     errno = 0;
     wchar_t *_endptr;
-    long long result = wcstoll(str, &_endptr, base);
+    long long result = std::wcstoll(str, &_endptr, base);
     while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
     if (!errno && *_endptr) {
         if (_endptr == str) {
@@ -731,7 +749,7 @@ unsigned long long fish_wcstoull(const wchar_t *str, const wchar_t **endptr, int
 
     errno = 0;
     wchar_t *_endptr;
-    unsigned long long result = wcstoull(str, &_endptr, base);
+    unsigned long long result = std::wcstoull(str, &_endptr, base);
     while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
     if (!errno && *_endptr) {
         if (_endptr == str) {
@@ -749,7 +767,7 @@ unsigned long long fish_wcstoull(const wchar_t *str, const wchar_t **endptr, int
 double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
     // The "fast path." If we're all ASCII and we fit inline, use strtod().
     char narrow[128];
-    size_t len = wcslen(str);
+    size_t len = std::wcslen(str);
     size_t len_plus_0 = 1 + len;
     auto is_digit = [](wchar_t c) { return '0' <= c && c <= '9'; };
     if (len_plus_0 <= sizeof narrow && std::all_of(str, str + len, is_digit)) {
@@ -810,6 +828,15 @@ file_id_t file_id_for_path(const wcstring &path) {
     return result;
 }
 
+file_id_t file_id_for_path(const std::string &path) {
+    file_id_t result = kInvalidFileID;
+    struct stat buf = {};
+    if (0 == stat(path.c_str(), &buf)) {
+        result = file_id_t::from_stat(buf);
+    }
+    return result;
+}
+
 bool file_id_t::operator==(const file_id_t &rhs) const { return this->compare_file_id(rhs) == 0; }
 
 bool file_id_t::operator!=(const file_id_t &rhs) const { return !(*this == rhs); }
